/**
 * See stratego_sglr_parse_table.c for information on the reasons for
 * using the "language" hacks.
 *
 * @author Martin Bravenboer
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <aterm2.h>
#include <srts/stratego.h>
#include <stratego-lib/stratego-lib-common.h>
#include <parse-table.h>
#include <rsrc-usage.h>
#include <MEPT-utils.h>

#include "stratego_sglr.h"
#include "stratego_sglr_internal.h"

/**
 * State
 */
static ATerm parse_error_term = NULL;
static ATerm none_term = NULL;
static ATerm true_term = NULL;
static ATerm false_term = NULL;

/**
 * Prototypes
 */
static char* read_text_from_stream(FILE* stream);

void STRSGLR_ensure_init(void)
{
  ATerm foo;

  static ATbool STRSGLR_init = ATfalse;
  if(STRSGLR_init)
    return;

  STRSGLR_init = ATtrue;

  PT_initMEPTApi();
  PT_initAsFix2Api();
  initErrorApi();

  SG_TOOLBUS_OFF();
  SG_OUTPUT_ON();
  foo = STRSGLR_set_default_config();

  none_term = (ATerm) ATmakeAppl(ATmakeAFun("None", 0, ATfalse));
  true_term = (ATerm) ATmakeAppl(ATmakeAFun("True", 0, ATfalse));
  false_term = (ATerm) ATmakeAppl(ATmakeAFun("False", 0, ATfalse));
  ATprotect(&none_term);
  ATprotect(&true_term);
  ATprotect(&false_term);
  ATprotect(&parse_error_term);
}

ATerm STRSGLR_set_default_config(void)
{
  STRSGLR_ensure_init();

  STRSGLR_set_default_filters();
  SG_ASFIX2_ON();
  SG_AMBIGUITY_ERROR_OFF();
  return (ATerm) ATempty;
}

/**
 * Parse a stream.
 *
 * sort is allowed to be None.
 */
ATerm STRSGLR_parse_stream_pt(StrCL onfail, ATerm stream_term, ATerm tbl_term, ATerm start_term, ATerm path_term, ATerm stop_term)
{
  ATerm result;
  int stop;
  parse_table* tbl = NULL;
  const char* start = NULL;
  const char* path = NULL;
  long pos;

  FILE* stream = stream_from_term(stream_term);
  if(stream == NULL)
    return NULL;
  pos = ftell(stream);

  STRSGLR_ensure_init();
  STRSGLR_clear_parse_error();

  tbl = parse_table_from_term(tbl_term);
  if(tbl == NULL)
    return NULL; // @todo: set the error term.

  if(AT_isString(start_term))
  {
    start = AT_getString(start_term);
  }
  else
  {
    if(start_term != none_term)
      return NULL;
  }

  if(stop_term == true_term)
  {
    stop = 1;
  }
  else
  {
    stop = 0;
  }

  if(AT_isString(path_term))
  {
    path = AT_getString(path_term);
  }
  else
  {
    return NULL;
  }

  result = SGparseStream(stream, (SGLR_ParseTable) tbl, start, path, stop);

  if(result != NULL)
  {
    if(SGisParseError(result))
    {
      parse_error_term = result;
      result = NULL;
    }
  }

  if(result == NULL) {
    fseek(stream, pos, SEEK_SET);
    cl_fun(onfail)(cl_sl(onfail), parse_error_term);
  }
  return result;
}

ATerm STRSGLR_get_parse_error(void)
{
  return parse_error_term;
}

ATerm STRSGLR_is_parse_tree(ATerm trm)
{
  if(SGisParseTree(trm))
  {
    return trm;
  }
  else
  {
    return NULL;
  }
}

ATerm STRSGLR_clear_parse_error(void)
{
  if(parse_error_term != NULL)
  {
    parse_error_term = NULL;
  }

  return (ATerm) ATempty;
}

ATerm STRSGLR_anno_position_info(ATerm pt_term, ATerm path_term)
{
  PT_ParseTree pt;
  const char* path;

  if(!AT_isString(path_term)) {
    return NULL;
  }

  path = AT_getString(path_term);
  pt = PT_ParseTreeFromTerm(pt_term);
  pt = PT_addParseTreePosInfoSome(path, pt, -1, ATfalse, ATfalse);

  return PT_ParseTreeToTerm(pt);
}

ATerm STRSGLR_implode_parsetree(ATerm t)
{
  return PT_implodeParseTree(
    PT_ParseTreeFromTerm(t),
    ATtrue, /* cons */
    ATtrue, /* layout */
    ATtrue, /* literals */
    ATtrue, /* injections */
    ATtrue, /* parsetree */
    ATtrue, /* implode lexicals */
    ATfalse, /* don't keep annotations */
    ATtrue, /* alt */
    ATtrue, /* seq */
    ATtrue, /* opt */
    ATfalse /* layout placeholder */
    );
}

static char* read_text_from_stream(FILE* stream)
{
  char* result;
  struct stat statbuf;
  int fd = fileno(stream);
  int c = 0;
  int position = 0;

  IF_STATISTICS(SG_Timer());

  if(fstat(fd, &statbuf) != 0)
    return NULL;

  if(S_ISREG(statbuf.st_mode))
  {
    size_t fsize = statbuf.st_size;
    result = malloc(fsize + 1);

    while((c = fgetc(stream)) != EOF)
    {
      result[position] = c;
      position++;
    }
  }
  else
  {
    int BUFSIZE = 65536;
    int bufsize = BUFSIZE;
    result = malloc(BUFSIZE);

    while((c = fgetc(stream)) != EOF)
    {
      if(position >= bufsize) {
        bufsize += BUFSIZE;
        result = realloc(result, bufsize);
      }

      result[position] = c;
      position++;
    }

    if(position >= bufsize)
      result = realloc(result, bufsize + 1);
  }

  result[position] = '\0';
  IF_STATISTICS(fprintf(SG_log(), "Reading stream of input: %.6f\n", SG_Timer()));

  return result;
}
