/* native-lazy.c: Native code of str-lazy.
* This file is part of Transformers.
* Copyright (C) 2005, 2006  EPITA Research and Development Laboratory.
*
* Transformers is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
* USA.
*/
#include <srts/stratego.h>
#include "native-lazy.h"


static Symbol invalid_sym;
static ATerm invalid_term;
static int debug_ambs;

static void init_module_constant_terms(void)
{
  ATprotect(&(invalid_term));
  invalid_term = (ATerm) ATmakeAppl(invalid_sym);
}

static
ATbool value_blob_destructor(ATermBlob);

static void init_module_constructors(void)
{
  invalid_sym = ATmakeSymbol("invalid", 0, ATfalse);
  ATprotectSymbol(invalid_sym);

#ifdef LAZY_DEBUG_AMBIGUITIES
  debug_ambs = 1;
#else
  debug_ambs = 0;
#endif

  ATregisterBlobDestructor(value_blob_destructor);
}

#include <srts/init-stratego-module.h>


#define MAGIC 0xA4B5C6D7

static
value_node_t *get_value_blob(ATerm a)
{
  ATermBlob b;
  value_blob_t *v;

  if (!ATisBlob(a))
  {
    ATfprintf(stderr, "get_value_blob: not a blob.\n");
    _fail(NULL);
  }
  b = (ATermBlob) a;

  if (ATgetBlobSize(b) != sizeof (value_blob_t))
  {
    ATfprintf(stderr, "get_value_blob: not a value blob (bad size).\n");
    ATfprintf(stderr, "Expecting %d, found %d\n", sizeof (value_blob_t), ATgetBlobSize(b));
    ATfprintf(stderr, "Magic code is 0x%08x, expecting: 0x%08x\n", *((int*)ATgetBlobData(b)), MAGIC);
    _fail(NULL);
  }
  v = ATgetBlobData(b);
  if (v->magic != MAGIC)
  {
    ATfprintf(stderr, "get_value_blob: not a value blob (bad magic).\n"
	      "Expecting 0x%08x, got 0x%08x\n", MAGIC, v->magic);
    _fail(NULL);
  }
  return v->node;
}

static struct str_frame fake_frame = { .parent = &fake_frame,
				       .vars = NULL,
				       .funs = NULL};

/**
 * Compute a value by checking its dependencies
 * All formulas represents differents ways to compute the value.
 * (ambiguities)
 *
 * + res_num + computed value +
 * ----------------------------
 * +    0    +  invalid_term  +
 * +    1    +  norm. result  +
 * +   > 1   +  invalid_term  + (2)
 *
 * (2): will return an error message: "too many valid results."
 *
 * res_num count the number of different results comming from
 * different branches.
 *
 */
static void compute(value_node_t *v)
{
  formula_list_t *l = v->formulas;
  unsigned int res_num = 0;

  ATprotect(&(v->value));
  v->value = invalid_term;
  while (l)
  {
    v->formulas = l->next;

    /*
    ** Warning: formulas can be registered by the strategy for the current
    ** attribute evaluation.
    */
    ATerm res = l->cur->strategy(&fake_frame, l->cur->input);

    if (res != NULL)
      if (!(match_cons(res, invalid_sym))) {
	if (ATisEqual(res, v->value) != ATtrue) {
	  if (res_num == 1) {
	    ATfprintf(stderr, "Has ambiguities in the result.\n");
	    ATfprintf(stderr, "Result 1: %t\n", v->value);
	    ATfprintf(stderr, "Result 2: %t\n", res);
	  }
	  else if (res_num > 1)
	    ATfprintf(stderr, "Result %d: %t\n", res_num+1, res);
	  if (res_num > 0)
	    v->value = ATmake("Ambiguous(<term>, <term>)", res, v->value);
	  else
	    v->value = res;
	  res_num++;
	}
	else
	  v->value = res;
      }
    ATunprotect(&(l->cur->input));
    free(l->cur);
    free(l);
    l = v->formulas;
  }
  if ((res_num > 1) && !debug_ambs)
    v->value = invalid_term;

  if (res_num <= 1)
    return;

  ATfprintf(stderr, "LAZY_compute: too many valid results (%i).\n", res_num);
}

ATerm LAZY_getv(ATerm t)
{
  value_node_t *value;

  value = get_value_blob(t);
  if (value == NULL)
    _fail(NULL);

  if (value->value == NULL)
    compute(value);

  if ((match_cons(value->value, invalid_sym)))
  {
    _fail(value->value);
  }

  return value->value;
}

/**
 * This function return the value iff it has already been computed with
 * getv. Otherwise this function fails.
 *
 * Caution: This function does not fail if the content of the attribute is
 * invalid!
 */
ATerm LAZY_getr(ATerm t)
{
  value_node_t *value;

  value = get_value_blob(t);
  if (value == NULL)
    _fail(NULL);

  if (value->value == NULL)
    _fail(NULL);

  return value->value;
}


static
ATerm make_value_blob(value_node_t *data)
{
  value_blob_t* b = malloc(sizeof (value_blob_t));

  if (b == NULL)
  {
    ATfprintf(stderr, "make_value_blob: allocation failed.\n");
    _fail(NULL);
  }

  b->magic = MAGIC;
  b->node = data;

  return (ATerm) ATmakeBlob(sizeof (value_blob_t), b);
}

static
ATbool value_blob_destructor(ATermBlob b)
{
  value_blob_t *v;
  value_node_t *node;
  formula_list_t *l, *old;

  if (ATgetBlobSize(b) != sizeof (value_blob_t))
    return ATfalse;
  v = ATgetBlobData(b);
  if (v->magic != MAGIC)
    return ATfalse;

  node = v->node;
  l = node->formulas;
  while (l) {
    ATunprotect(&(l->cur->input));
    free(l->cur);
    old = l;
    l = l->next;
    free(old);
  }
  if (node->value != NULL)
    ATunprotect(&(node->value));
  free(node);
  free(v);

  return ATtrue;
}


ATerm LAZY_new_value(void)
{
  value_node_t *node = calloc(1, sizeof(value_node_t));

  if (node == NULL)
  {
    ATfprintf(stderr, "LAZY_new_value: allocation failed.\n");
    _fail(NULL);
  }

  node->value = NULL;
  node->formulas = NULL;

  return make_value_blob(node);
}

ATerm LAZY_formula(StrCL s,
		   ATerm result, ATerm input)
{
  formula_node_t *formula = calloc(1, sizeof(formula_node_t));
  value_node_t *value;
  formula_list_t *list;

  if (formula == NULL)
  {
    ATfprintf(stderr, "LAZY_formula: allocation failed.\n");
    _fail(input);
  }

  ATprotect(&(formula->input));

  formula->strategy = cl_fun(s);
  formula->input = input;

  value = get_value_blob(result);
  if (value == NULL)
    _fail(NULL);

  list = value->formulas;
  value->formulas = malloc(sizeof(value_list_t));
  if (value->formulas == NULL)
  {
    ATfprintf(stderr, "LAZY_formula: allocation failed.\n");
    value->formulas = list;
    _fail(result);
  }
  value->formulas->cur = formula;
  value->formulas->next = list;

  return input;
}
